@@ -6,8 +6,13 @@ import socket
 import time  # noqa: F401
 import gc  # noqa: F401
 
-from threading import Lock, Condition
-from rpyc.lib import spawn, Timeout, get_methods, get_id_pack
+import collections
+import concurrent.futures as c_futures
+import os
+import threading
+
+from threading import Lock, Condition, RLock
+from rpyc.lib import spawn, Timeout, get_methods, get_id_pack, hasattr_static
 from rpyc.lib.compat import pickle, next, maxint, select_error, acquire_lock  # noqa: F401
 from rpyc.lib.colls import WeakValueDict, RefCountingColl
 from rpyc.core import consts, brine, vinegar, netref
@@ -19,6 +24,7 @@ class PingError(Exception):
     pass
 
 
+UNBOUND_THREAD_ID = 0  # Used when the message is being sent but the thread is not bound yet.
 DEFAULT_CONFIG = dict(
     # ATTRIBUTES
     allow_safe_attrs=True,
@@ -62,6 +68,7 @@ DEFAULT_CONFIG = dict(
     sync_request_timeout=30,
     before_closed=None,
     close_catchall=False,
+    bind_threads=os.environ.get('RPYC_BIND_THREADS') == 'true',
 )
 """
 The default configuration dictionary of the protocol. You can override these parameters
@@ -109,7 +116,7 @@ Parameter                                Default value     Description
 
 ``connid``                               ``None``          **Runtime**: the RPyC connection ID (used
                                                            mainly for debugging purposes)
-``credentials``                          ``None``          **Runtime**: the credentails object that was returned
+``credentials``                          ``None``          **Runtime**: the credentials object that was returned
                                                            by the server's :ref:`authenticator <api-authenticators>`
                                                            or ``None``
 ``endpoints``                            ``None``          **Runtime**: The connection's endpoints. This is a tuple
@@ -119,6 +126,9 @@ Parameter                                Default value     Description
                                                            do no have this configuration option set.
 
 ``sync_request_timeout``                 ``30``            Default timeout for waiting results
+``bind_threads``                         ``False``         Whether to restrict request/reply by thread (experimental).
+                                                           The default value is False. Setting the environment variable
+                                                           `RPYC_BIND_THREADS` to `"true"` will enable this feature.
 =======================================  ================  =====================================================
 """
 
@@ -129,6 +139,11 @@ _connection_id_generator = itertools.count(1)
 class Connection(object):
     """The RPyC *connection* (AKA *protocol*).
 
+    Objects referenced over the connection are either local or remote. This class retains a strong reference to
+    local objects that is deleted when the reference count is zero. Remote/proxied objects have a life-cycle
+    controlled by a different address space. Since garbage collection is handled on the remote end, a weak reference
+    is used for netrefs.
+
     :param root: the :class:`~rpyc.core.service.Service` object to expose
     :param channel: the :class:`~rpyc.core.channel.Channel` over which messages are passed
     :param config: the connection's configuration dict (overriding parameters
@@ -145,9 +160,9 @@ class Connection(object):
         self._HANDLERS = self._request_handlers()
         self._channel = channel
         self._seqcounter = itertools.count()
-        self._recvlock = Lock()
+        self._recvlock = RLock()  # AsyncResult implementation means that synchronous requests have multiple acquires
         self._sendlock = Lock()
-        self._recv_event = Condition()
+        self._recv_event = Condition()  # TODO: why not simply timeout? why not associate w/ recvlock? explain/redesign
         self._request_callbacks = {}
         self._local_objects = RefCountingColl()
         self._last_traceback = None
@@ -157,6 +172,15 @@ class Connection(object):
         self._send_queue = []
         self._local_root = root
         self._closed = False
+        # Settings for bind_threads
+        self._bind_threads = self._config['bind_threads']
+        self._threads = None
+        if self._bind_threads:
+            self._lock = threading.Lock()
+            self._threads = {}
+            self._receiving = False
+            self._thread_pool = []
+            self._thread_pool_executor = c_futures.ThreadPoolExecutor()
 
     def __del__(self):
         self.close()
@@ -187,6 +211,8 @@ class Connection(object):
         # self._seqcounter = None
         # self._config.clear()
         del self._HANDLERS
+        if self._bind_threads:
+            self._thread_pool_executor.shutdown(wait=False)  # TODO where?
 
     def close(self):  # IO
         """closes the connection, releasing all held resources"""
@@ -235,6 +261,15 @@ class Connection(object):
 
     def _send(self, msg, seq, args):  # IO
         data = brine.dump((msg, seq, args))
+        if self._bind_threads:
+            this_thread = self._get_thread()
+            data = brine.I8I8.pack(this_thread.id, this_thread._remote_thread_id) + data
+            if msg == consts.MSG_REQUEST:
+                this_thread._occupation_count += 1
+            else:
+                this_thread._occupation_count -= 1
+                if this_thread._occupation_count == 0:
+                    this_thread._remote_thread_id = UNBOUND_THREAD_ID
         # GC might run while sending data
         # if so, a BaseNetref.__del__ might be called
         # BaseNetref.__del__ must call asyncreq,
@@ -359,15 +394,23 @@ class Connection(object):
     def _dispatch(self, data):  # serving---dispatch?
         msg, seq, args = brine.load(data)
         if msg == consts.MSG_REQUEST:
+            if self._bind_threads:
+                self._get_thread()._occupation_count += 1
             self._dispatch_request(seq, args)
-        elif msg == consts.MSG_REPLY:
-            obj = self._unbox(args)
-            self._seq_request_callback(msg, seq, False, obj)
-        elif msg == consts.MSG_EXCEPTION:
-            obj = self._unbox_exc(args)
-            self._seq_request_callback(msg, seq, True, obj)
         else:
-            raise ValueError(f"invalid message type: {msg!r}")
+            if self._bind_threads:
+                this_thread = self._get_thread()
+                this_thread._occupation_count -= 1
+                if this_thread._occupation_count == 0:
+                    this_thread._remote_thread_id = UNBOUND_THREAD_ID
+            if msg == consts.MSG_REPLY:
+                obj = self._unbox(args)
+                self._seq_request_callback(msg, seq, False, obj)
+            elif msg == consts.MSG_EXCEPTION:
+                obj = self._unbox_exc(args)
+                self._seq_request_callback(msg, seq, True, obj)
+            else:
+                raise ValueError(f"invalid message type: {msg!r}")
 
     def serve(self, timeout=1, wait_for_lock=True):  # serving
         """Serves a single request or reply that arrives within the given
@@ -375,26 +418,225 @@ class Connection(object):
         might trigger multiple (nested) requests, thus this function may be
         reentrant.
 
-        :returns: ``True`` if a request or reply were received, ``False``
-                  otherwise.
+        :returns: ``True`` if a request or reply were received, ``False`` otherwise.
         """
         timeout = Timeout(timeout)
+        if self._bind_threads:
+            return self._serve_bound(timeout, wait_for_lock)
         with self._recv_event:
+            # Exit early if we cannot acquire the recvlock
             if not self._recvlock.acquire(False):
-                return wait_for_lock and self._recv_event.wait(timeout.timeleft())
+                if wait_for_lock:
+                    # Wait condition for recvlock release; recvlock is not underlying lock for condition
+                    return self._recv_event.wait(timeout.timeleft())
+                else:
+                    return False
+        # Assume the receive rlock is acquired and incremented
+        # We must release once BEFORE dispatch, dispatch any data, and THEN notify all (see issue #527 and #449)
         try:
+            data = None  # Ensure data is initialized
             data = self._channel.poll(timeout) and self._channel.recv()
-            if not data:
-                return False
-        except EOFError:
-            self.close()
-            raise
-        finally:
+        except Exception as exc:
             self._recvlock.release()
+            if isinstance(exc, EOFError):
+                self.close()  # sends close async request
+            raise
+        else:
+            self._recvlock.release()
+            if data:
+                self._dispatch(data)  # Dispatch will unbox, invoke callbacks, etc.
+                return True
+            else:
+                return False
+        finally:
             with self._recv_event:
                 self._recv_event.notify_all()
-        self._dispatch(data)
-        return True
+
+    def _serve_bound(self, timeout, wait_for_lock):
+        """Serves messages like `serve` with the added benefit of making request/reply thread bound.
+        - Experimental functionality `RPYC_BIND_THREADS`
+
+        The first 8 bytes indicate the sending thread ID and intended recipient ID. When the recipient
+        thread ID is not the thread that received the data, the remote thread ID and message are appended
+        to the intended threads `_deque` and `_event` is set.
+
+        :returns: ``True`` if a request or reply were received, ``False`` otherwise.
+        """
+        this_thread = self._get_thread()
+        wait = False
+
+        with self._lock:
+            message_available = this_thread._event.is_set() and len(this_thread._deque) != 0
+
+            if message_available:
+                remote_thread_id, message = this_thread._deque.popleft()
+                if len(this_thread._deque) == 0:
+                    this_thread._event.clear()
+
+            else:
+                if self._receiving:  # enter pool
+                    self._thread_pool.append(this_thread)
+                    wait = True
+
+                else:
+                    self._receiving = True
+
+        if message_available:  # just process
+            this_thread._remote_thread_id = remote_thread_id
+            self._dispatch(message)
+            return True
+
+        if wait:
+            while True:
+                if wait_for_lock:
+                    this_thread._event.wait(timeout.timeleft())
+
+                with self._lock:
+                    if this_thread._event.is_set():
+                        message_available = len(this_thread._deque) != 0
+
+                        if message_available:
+                            remote_thread_id, message = this_thread._deque.popleft()
+                            if len(this_thread._deque) == 0:
+                                this_thread._event.clear()
+
+                        else:
+                            this_thread._event.clear()
+
+                            if self._receiving:  # another thread was faster
+                                continue
+
+                            self._receiving = True
+
+                        self._thread_pool.remove(this_thread)  # leave pool
+                        break
+
+                    else:  # timeout
+                        return False
+
+            if message_available:
+                this_thread._remote_thread_id = remote_thread_id
+                self._dispatch(message)
+                return True
+
+        while True:
+            # from upstream
+            try:
+                message = self._channel.poll(timeout) and self._channel.recv()
+
+            except Exception as exception:
+                if isinstance(exception, EOFError):
+                    self.close()  # sends close async request
+
+                with self._lock:
+                    self._receiving = False
+
+                    for thread in self._thread_pool:
+                        thread._event.set()
+                        break
+
+                raise
+
+            if not message:  # timeout; from upstream
+                with self._lock:
+                    for thread in self._thread_pool:
+                        if not thread._event.is_set():
+                            self._receiving = False
+                            thread._event.set()
+                            break
+
+                    else:  # stop receiving
+                        self._receiving = False
+
+                return False
+
+            remote_thread_id, local_thread_id = brine.I8I8.unpack(message[:16])
+            message = message[16:]
+
+            this = False
+
+            if local_thread_id == UNBOUND_THREAD_ID and this_thread._occupation_count != 0:
+                # Message is not meant for this thread. Use a thread that is not occupied or have the pool create a new one.
+                # TODO: reusing threads may be problematic if occupation being zero is wrong...
+                new = False
+                with self._lock:
+                    for thread in self._thread_pool:
+                        if thread._occupation_count == 0 and not thread._event.is_set():
+                            thread._deque.append((remote_thread_id, message))
+                            thread._event.set()
+                            break
+
+                    else:
+                        new = True
+
+                if new:
+                    self._thread_pool_executor.submit(self._serve_temporary, remote_thread_id, message)
+
+            elif local_thread_id in {UNBOUND_THREAD_ID, this_thread.id}:
+                # Of course, the message is for this thread if equal. When id is UNBOUND_THREAD_ID,
+                # we deduce that occupation count is 0 from the previous if condition. So, set this True.
+                this = True
+            else:
+                # Otherwise, message was meant for another thread.
+                thread = self._get_thread(id=local_thread_id)
+                with self._lock:
+                    thread._deque.append((remote_thread_id, message))
+                    thread._event.set()
+
+            if this:
+                with self._lock:
+                    for thread in self._thread_pool:
+                        if not thread._event.is_set():
+                            self._receiving = False
+                            thread._event.set()
+                            break
+
+                    else:  # stop receiving
+                        self._receiving = False
+
+                this_thread._remote_thread_id = remote_thread_id
+                self._dispatch(message)
+                return True
+
+    def _serve_temporary(self, remote_thread_id, message):
+        """Callable that is used to schedule serve as a new thread
+        - Experimental functionality `RPYC_BIND_THREADS`
+
+        :returns: None
+        """
+        thread = self._get_thread()
+        thread._deque.append((remote_thread_id, message))
+        thread._event.set()
+
+        # from upstream
+        try:
+            while not self.closed:
+                self.serve(None)
+
+                if thread._occupation_count == 0:
+                    break
+
+        except (socket.error, select_error, IOError):
+            if not self.closed:
+                raise
+        except EOFError:
+            pass
+
+    def _get_thread(self, id=None):
+        """Get internal thread information for current thread for ID, when None use current thread.
+        - Experimental functionality `RPYC_BIND_THREADS`
+
+        :returns: _Thread
+        """
+        if id is None:
+            id = threading.get_ident()
+
+        thread = self._threads.get(id)
+        if thread is None:
+            thread = _Thread(id)
+            self._threads[id] = thread
+
+        return thread
 
     def poll(self, timeout=0):  # serving
         """Serves a single transaction, should one arrives in the given
@@ -463,14 +705,17 @@ class Connection(object):
             pass
         return at_least_once
 
-    def sync_request(self, handler, *args):  # serving
+    def sync_request(self, handler, *args):
         """requests, sends a synchronous request (waits for the reply to arrive)
 
         :raises: any exception that the requets may be generated
         :returns: the result of the request
         """
         timeout = self._config["sync_request_timeout"]
-        return self.async_request(handler, *args, timeout=timeout).value
+        _async_res = self.async_request(handler, *args, timeout=timeout)
+        # _async_res is an instance of AsyncResult, the value property invokes Connection.serve via AsyncResult.wait
+        # So, the _recvlock can be acquired multiple times by the owning thread and warrants the use of RLock
+        return _async_res.value
 
     def _async_request(self, handler, args=(), callback=(lambda a, b: None)):  # serving
         seq = self._get_seq_id()
@@ -515,7 +760,7 @@ class Connection(object):
         plain |= config["allow_exposed_attrs"] and name.startswith(prefix)
         plain |= config["allow_safe_attrs"] and name in config["safe_attrs"]
         plain |= config["allow_public_attrs"] and not name.startswith("_")
-        has_exposed = prefix and hasattr(obj, prefix + name)
+        has_exposed = prefix and (hasattr(obj, prefix + name) or hasattr_static(obj, prefix + name))
         if plain and (not has_exposed or hasattr(obj, name)):
             return name
         if has_exposed:
@@ -669,3 +914,17 @@ class Connection(object):
                 stop = maxint
             getslice = self._handle_getattr(obj, fallback)
             return getslice(start, stop, *args)
+
+
+class _Thread:
+    """Internal thread information for the RPYC protocol used for thread binding."""
+
+    def __init__(self, id):
+        super().__init__()
+
+        self.id = id
+
+        self._remote_thread_id = UNBOUND_THREAD_ID
+        self._occupation_count = 0
+        self._event = threading.Event()
+        self._deque = collections.deque()
