@@ -80,7 +80,8 @@ class Server(object):
             else:
                 family = socket.AF_INET
             self.listener = socket.socket(family, socket.SOCK_STREAM)
-            address = socket.getaddrinfo(hostname, port, family=family, type=socket.SOCK_STREAM, proto=socket.IPPROTO_TCP, flags=socket.AI_PASSIVE)[0][-1]
+            address = socket.getaddrinfo(hostname, port, family=family, type=socket.SOCK_STREAM,
+                                         proto=socket.IPPROTO_TCP, flags=socket.AI_PASSIVE)[0][-1]
 
             if reuse_addr and sys.platform != "win32":
                 # warning: reuseaddr is not what you'd expect on windows!
@@ -108,7 +109,6 @@ class Server(object):
         also unregisters from the registry server"""
         if self._closed:
             return
-        self._closed = True
         self.active = False
         if self.auto_register:
             try:
@@ -128,6 +128,7 @@ class Server(object):
                 pass
             c.close()
         self.clients.clear()
+        self._closed = True
 
     def fileno(self):
         """returns the listener socket's file descriptor"""
@@ -341,7 +342,7 @@ class ThreadPoolServer(Server):
         self.workers = []
         for i in range(self.nbthreads):
             t = spawn(self._serve_clients)
-            t.setName(f"Worker{i}")
+            t.name = f"Worker{i}"
             self.workers.append(t)
         # setup a thread for polling inactive connections
         self.polling_thread = spawn(self._poll_inactive_clients)
